
from html import unescape
from typing import List, Tuple
from prompt_toolkit import HTML
from prompt_toolkit.application import Application
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout import Layout
from prompt_toolkit.layout.containers import HSplit
from prompt_toolkit.widgets import TextArea, SearchToolbar,Label
from prompt_toolkit.key_binding.bindings import search
from prompt_toolkit.key_binding import ConditionalKeyBindings
from prompt_toolkit.filters.app import is_searching
from prompt_toolkit.filters.base import Condition
from prompt_toolkit.styles import Style

from dAngr.utils.utils import remove_xml_tags

class Less:

    def __init__(self):
        self.inner:TextArea = None # type: ignore
        self.search_field:SearchToolbar = None # type: ignore
        self.app:Application = None # type: ignore
        pass
            
    def show_less(self, txts:List[Tuple[str,Style|None]]):
        self.app = Application(
            layout=self._get_less_layout(texts=txts),
            key_bindings=self._get_less_key_bindings(),
            full_screen=True,
            mouse_support=True,
        )
        self.app.run()

    def _get_less_layout(self, texts:List[Tuple[str,Style|None]]):
        self.search_field = SearchToolbar()
        info = Label(HTML("<darkgray>Press 'ctrl-d' to exit, 'ctrl-f' to forward search, 'ctrl-r' for reverse search</darkgray>"))
        # TODO: display formatted text with styles
        txts = [unescape(remove_xml_tags(t[0])) for t in texts]
        self.inner = TextArea("\n".join(txts),scrollbar=True,read_only=True, search_field=self.search_field, multiline=True, wrap_lines=True, focus_on_click=True)
        return Layout(
                HSplit([ 
                    self.inner, 
                    self.search_field, 
                    info
                ])
                )

    def _get_less_key_bindings(self):
        kb = KeyBindings()
        
        @kb.add('c-d')
        def _(event):
            event.app.exit()

        handle = kb.add


        @Condition
        def search_buffer_is_empty() -> bool:
            "Returns True when the search buffer is empty."
            return self.app.current_buffer.text == ""

        handle("c-f")(search.start_forward_incremental_search)
        handle("c-r")(search.start_reverse_incremental_search)

        # Apply the search. (At the / or ? prompt.)
        handle("enter", filter=is_searching)(search.accept_search)

        handle("c-f", filter=is_searching)(search.forward_incremental_search)
        handle("c-s", filter=is_searching)(search.reverse_incremental_search)


        return ConditionalKeyBindings(kb)


import importlib
import importlib.util
import inspect
import os
import types
from typing import Any, Set, List, Tuple
from angr import Project, SimCC, SimProcedure, SimState, SimulationManager, types, knowledge_plugins



from dAngr.angr_ext.std_tracker import StdTracker
from dAngr.exceptions import DebuggerCommandError, InvalidArgumentError, FileNotFoundError

def evaluate_symbolic_string(symbolic_str, solver, length):
    result = []
    for i in range(length):
        byte = symbolic_str.get_byte(i)
        result.append(solver.eval(byte, cast_to=int))
    return bytes(result)

def create_entry_state(project:Project, 
                       entry_point:int|Tuple[str,types.SimTypeFunction,SimCC,List[Any]]|None= None, 
                       default_state_options:Set[str]=set(), state = None,
                       veritesting:bool=False)-> Tuple[SimulationManager,SimState]:
        if not state:
            if entry_point is None:
                state = project.factory.entry_state(add_options=default_state_options)
            elif isinstance(entry_point,int):
                state = project.factory.blank_state(addr=entry_point, add_options=default_state_options)
            else:
                name,prototype,cc,arguments = entry_point
                addr = get_function_address(project,name)
                state = project.factory.call_state( 
                    addr,
                    *arguments,
                    prototype=prototype,
                    cc=cc,
                    base_state=None,
                    ret_addr=project.simos.return_deadend, 
                    toc=None,
                    add_options= default_state_options,
                    remove_options=None,
                )
        state.register_plugin('stdout_tracker', StdTracker())
        simgr = project.factory.simulation_manager(state, veritesting=veritesting)
        return simgr,state

def get_function_address(project, function_name):
    for symbol in project.loader.symbols: 
        if symbol.name == function_name:
            return symbol.rebased_addr
    if f := project.kb.functions.function(name=function_name):
        return f.addr
    return None


def convert_string(sim_type, value):
    if isinstance(sim_type, types.SimTypeInt):
        return int(value,0)
    elif isinstance(sim_type, types.SimTypePointer) :
        return eval(value)
    elif isinstance(sim_type, types.SimTypeDouble) or isinstance(sim_type, types.SimTypeFloat):
        return float(value)
    else:
        raise InvalidArgumentError(f"arg_type {sim_type} not implemented")
    
def get_function_by_addr(proj,addr) -> knowledge_plugins.functions.function.Function | None:
    return proj.kb.functions.ceiling_func(addr)

def get_function_by_name(proj,name) -> knowledge_plugins.functions.function.Function | None:
    function_list = list(proj.kb.functions.get_by_name(name=name))
    if not function_list:
        return None
    
    return function_list[0]

def get_bb_end_address(state)->int:
    bb = state.project.factory.block(state.addr)
    if not bb.instruction_addrs:
        return state.addr
    return bb.instruction_addrs[-1]

def load_module_from_file(file_path):
    # Extract filename without extension
    module_name = os.path.splitext(os.path.basename(file_path))[0]

    # Create spec for the module
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    if spec is None:
        raise FileNotFoundError(f"File '{file_path}' not found.")
    if spec.loader is None:
        raise DebuggerCommandError(f"Failed to load '{file_path}'.")
    # Load the module
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    return module

def hook_simprocedures(project, module):
    hooks = []
    for obj_name in dir(module):
        obj = getattr(module, obj_name)
        if inspect.isclass(obj) and issubclass(obj, SimProcedure):
            hooks.append(obj_name)
            project.hook_symbol(obj_name, obj())
    return hooks





from abc import abstractmethod


class Connection:
    def __init__(self):
        pass
    
    @abstractmethod
    def send_result(self, data,newline=True, style=None)->None:
        raise NotImplementedError
    
    @abstractmethod
    def send_info(self, data, style=None)->None:
        raise NotImplementedError

    @abstractmethod
    def send_output(self, data, style=None)->None:
        raise NotImplementedError

    @abstractmethod
    def send_error(self, data, style=None)->None:
        raise NotImplementedError
    
    @abstractmethod
    def send_warning(self, data, style=None)->None:
        raise NotImplementedError




import os
import re

class ScriptProcessor:
    def __init__(self, script_path):
        self.script_path = script_path
        self.curdir = os.path.realpath(os.curdir)
    
    #destructor
    def __del__(self):
        os.chdir(self.curdir)
    
    def is_markdown_file(self):
        return self.script_path.lower().endswith(('.md', '.markdown'))

    def process_file(self):
        if not os.path.exists(self.script_path):
            raise FileNotFoundError(f"File '{self.script_path}' not found.")
        with open(self.script_path, 'r') as f:
            if os.path.dirname(self.script_path):
                os.chdir(os.path.dirname(self.script_path))
            
            if self.is_markdown_file():
                yield from self.process_markdown(f)
            else:
                yield from self.process_text(f)

    def process_text(self, file_obj, until=lambda line: False):
        # if definition or control flow (end with :), read until back to 0 indentation
        l = ""
        line = None
        stack = False
        while True:
            if not line:
                line = next(file_obj, None)
            if line is None:
                break
            if until(line):
                if l:
                    yield l
                break
            if line.strip() == "":
                line = None
                continue
            if not stack and line.strip().endswith(":"):
                stack = True
            elif line.find(line.lstrip()) == 0:
                stack = False
            if stack:
                l += "\n" + line.rstrip() if l else line.rstrip()
                line = None
            else:
                if l:
                    yield l
                    l = ""
                else:
                    yield line.strip()
                    line = None


    def process_markdown(self, file_obj):
        for line in file_obj:
            line = line.rstrip()
            if line.strip() == "":
                continue
            # Handle inline code (between single backticks)
            inline_code_matches = re.findall(r'`([^`]+)`', line)
            for code in inline_code_matches:
                yield f"{code}".strip()

            # Handle code blocks (between triple backticks or triple single quotes)
            prefix = line[:3]
            if prefix in ['```', "[[["]:
                postfix = '```' if prefix == '```' else ']]]'
                # Yield the collected code block lines
                for l in self.process_text(file_obj, lambda line:line.startswith(postfix)):
                    if l.strip():
                        yield l.strip()





# Assuming all command classes are in debugger_commands.py
import os
import threading
from typing import Callable, Dict, List, Tuple, cast

import angr


import archinfo
import claripy
from prompt_toolkit.styles import Style

from dAngr.angr_ext.step_handler import StepHandler, StopReason
from dAngr.angr_ext.debugger import Debugger

from dAngr.cli.debugger_commands import *
from dAngr.cli.debugger_commands.base import BuiltinFunctionDefinition
from dAngr.cli.filters import Filter, FilterList
from dAngr.cli.grammar.parser import parse_input, validate_input
from dAngr.cli.grammar.execution_context import ExecutionContext
from dAngr.cli.grammar.expressions import Object

from dAngr.exceptions import CommandError
from dAngr.exceptions.DebuggerCommandError import DebuggerCommandError

from dAngr.utils.utils import DataType, get_union_members
from .cli_connection import CliConnection
from .debugger_commands import *

from dAngr.utils.loggers import get_logger
log = get_logger(__name__)


def _auto_register_commands() -> Dict[str, BuiltinFunctionDefinition]:
    commands:Dict[str, BuiltinFunctionDefinition] = {}
    for name, obj in globals().items():
        if isinstance(obj, type) and issubclass(obj, BaseCommand) and obj is not BaseCommand:
            specs = getattr(obj, "__cmd_specs__", None)
            if specs:
                for command in specs:
                    spec = specs[command]
                    # Add the command to the list of available commands
                    commands[spec.name] = spec
    return commands
DEBUGGER_COMMANDS = _auto_register_commands()

EMPTY = "\t"

class dAngrExecutionContext(ExecutionContext):
    def __init__(self, debugger:Debugger, commands:Dict[str, BuiltinFunctionDefinition]):
        super().__init__()
        self._debugger = debugger
        #add commands to context
        self._definitions.update(commands)
    
    def clone(self):
        e = dAngrExecutionContext(self._debugger, cast(Dict[str, BuiltinFunctionDefinition], self._definitions))
        e._variables = self._variables.copy()
        e._parent = self
        return e

    @property
    def debugger(self)->Debugger:
        return self._debugger

class CommandLineDebugger(Debugger,StepHandler):
    def __init__(self, *args):
        conn = args[0]
        Debugger.__init__(self, conn)
        self._breakpoints:FilterList = FilterList([])
        self._exclusions:List[Filter] = [] #TODO: check if we can use FilterList
        self.http_server = None
        self.http_thread = None
        self.context:dAngrExecutionContext = dAngrExecutionContext(self, DEBUGGER_COMMANDS)

    def reset_state(self):
        self.http_server = None
        self.http_thread = None
        return super().reset_state()

    def __del__(self):
        if self.http_server is not None:
            self.http_server.shutdown()
            self.http_server.server_close()
        if self.http_thread is not None:
            self.http_thread.join()
    
    @property
    def breakpoints(self)->FilterList:
        self.throw_if_not_initialized()
        return self._breakpoints
    
    @property
    def exclusions(self)->List[Filter]:
        self.throw_if_not_initialized()
        return self._exclusions

    # step callback methods
    def handle_output(self, output:str):
        self.conn.send_output(f"{output}")
    
    def handle_step(self,reason:StopReason, state:angr.SimState|None):
        if reason == StopReason.TERMINATE:
            self.conn.send_info(f"Terminated.") # type: ignore
        elif state is None:
            self.conn.send_warning(f"Stopped with unknown reason.")
        elif reason == StopReason.STEP:
            self.conn.send_info(f"Stepped to: {hex(state.addr)}.") # type: ignore
        elif reason == StopReason.BREAKPOINT:
            for f in self.breakpoints.get_matching_filter(state):
                self.conn.send_info(f"Break: {f}.")
        else:
            self.conn.send_warning(f"Stopped with unknown reason at: {hex(start)}.") # type: ignore


    def validate_input(self, command):
        return validate_input(command)
        
    # command methods
    def handle(self, command:str, raise_error:bool = True):
        try:
            if not command:
                return True
            elif command == "exit":
                return False
            # command = command.strip()
            script = parse_input(command, debugger=self)
            if not script:
                self.conn.send_info("No command entered.")
                return True
            r = script(self.context)
            if r is not None:
                self.conn.send_result(r, True)
            cast(CliConnection,self.conn).clear_output()
            return True
        except CommandError as e:
            if raise_error:
                raise e
            else:
                self.conn.send_error(e)
                return True
            # raise e
        
    def launch_file_server(self):
        base_path = "/tmp/web" 
        # set file_index to 1 + nr of files in path
        if self.http_server is None:
            if not os.path.exists(base_path):
                os.makedirs(base_path)
            else:
                for f in os.listdir(base_path):
                    os.remove(os.path.join(base_path, f))
            import http.server
            class Handler(http.server.SimpleHTTPRequestHandler):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs, directory=base_path)

            import socketserver
            PORT = 8000
            handler = Handler
            self.http_server = socketserver.TCPServer(("127.0.0.1", PORT), handler)
            #serve files in a separate thread
            self.http_thread = threading.Thread(target=self.http_server.serve_forever).start()
        return base_path


    def get_command_spec(self, command:str)->BuiltinFunctionDefinition|None:
        return DEBUGGER_COMMANDS.get(command, None)


    def list_commands_data(self, withArgs:bool = False):
        table_data = [[],["<title>Available commands:</title>"]]
        package = ''
        # get a list of keys from the dictionary sorted based on package and name
        list_of_keys = sorted((k for k in DEBUGGER_COMMANDS.keys() if DEBUGGER_COMMANDS[k].short_name!= k ), key=lambda x: (DEBUGGER_COMMANDS[x].package, x))
        for command in list_of_keys:
            spec = DEBUGGER_COMMANDS[command]
            if command != spec.name:
                continue
            # get module base name if different from the previous one
            if package != spec.package:
                package = spec.package
                table_data.append([EMPTY, f"<package>{package}</package>"])

            table_data.append([EMPTY, EMPTY,f"<command>{command} <shortcmd>({spec.short_name})</shortcmd></command>:"])
            if not withArgs:
                # get first line of info and append ... if there are more lines
                first, sec = spec.description.split('\n', 1) if '\n' in spec.description else (spec.description, '')
                if sec:
                    first += "..."
                if len(first.strip("..."))>50:
                    first = first[:50] + "..."
                
                table_data.append([EMPTY, EMPTY, EMPTY, f"<info>{first}</info>"])
            else:
                dt, _ =self.list_args_table(spec,[EMPTY,EMPTY,EMPTY])
                table_data.extend(dt)
            table_data.append([])
        style = Style.from_dict({
            'title': '',
            'package': 'yellow italic',
            'command': 'blue bold',
            'shortcmd': 'italic',
            'info': 'darkgray',
        })
        return table_data, style
    

    def list_args_table(self, spec :BuiltinFunctionDefinition,indent = []) -> Tuple[List[List[str]],Style]:
        style = Style.from_dict({
            'title': '',
            'command': 'blue bold',
            'info': '',
            'argument': 'green',
            'arg_info': 'darkgray italic',
            'extra_info': 'gray italic',
        })

        table_data = []
        for l in spec.description.split("\n"):
            table_data.append([EMPTY, f"<info>{l}</info>"])
                
        # command structure
        # print command name, followed by required args separated by comma, then optional args in square brackets
        required = spec.required_arguments
        optional = spec.optional_arguments
        req = " " + " ".join([f"<argument>{a.name}</argument>" for a in required])
        opt = " " + " ".join([f"<argument>[{a.name}]</argument>" for a in optional])
        table_data.append([EMPTY, f"- Usage: <command>{spec.name}{req.rstrip()}{opt.rstrip()}</command>"])
        table_data.append([EMPTY, f"- Short name: <command>{spec.short_name}</command>"])
        if spec.args:
            if required:
                table_data.append([EMPTY, "<info>- Arguments:</info>"])
            for a in required:                
                if a.dtype:
                    if members:= get_union_members(a.dtype):
                        tp = " or ".join([m.__name__ for m in members])
                    else:
                        tp = a.dtype.__name__
                else: tp ="any"
                table_data.append([EMPTY, EMPTY, f"<argument>{a.name}</argument> <arg_info>({tp})</arg_info>"])
                if a.description:
                    table_data.append([EMPTY, EMPTY, EMPTY, f"<arg_info>{a.description}</arg_info>"])
            if optional:
                table_data.append([EMPTY, "<info>Optional Arguments:</info>"]) 
                for a in optional:
                    if a.dtype:
                        if members:= get_union_members(a.dtype):
                            tp = " or ".join([m.__name__ for m in members])
                        else:
                            tp = a.dtype.__name__
                    else: tp ="any"
                    table_data.append([EMPTY, EMPTY, f"<argument>{a.name}</argument> <arg_info>({tp})\t</arg_info>"])
                    if a.dtype:
                        table_data.append([EMPTY, EMPTY, EMPTY, f"{a.description}"])

        if spec.example:
            table_data.append([EMPTY, "<info>Example:</info>"])
            for l in spec.example.split("\n"):
                table_data.append([EMPTY, EMPTY, f"<extra_info>{l}</extra_info>"])
        #insert indent to each row:
        for i in range(len(table_data)):
            table_data[i] = indent + table_data[i]
        return table_data, style


    def render_argument(self, value:int|bytes|str|Object, make_concrete:bool ):
        val = None
        if isinstance(value, int):
            val = DataType._to_bytes(value, archinfo.get_host_arch())
            if not make_concrete:
                val = claripy.BVV(value, len(val)*8)
        elif isinstance(value, bytes):
            if not make_concrete:
                val = claripy.BVV(value)
            pass
        elif isinstance(value, str):
            val = DataType._to_bytes(value, archinfo.get_host_arch())
            if not make_concrete:
                val = claripy.BVV(val)
        elif isinstance(value, bool):
            if not make_concrete:
                val = claripy.BVV(1 if value else 0, 1)
        elif isinstance(value, Object):
            val = value(context=self.context)
        if val == None:
            raise DebuggerCommandError(f"Could not convert value {value}")
        return val
    
    def list_commands(self, withArgs:bool = False):
        table_data, style = self.list_commands_data(withArgs)

        # Create HTML table
        html_table = ""
        for row in table_data:
            html_table += "\t"
            for cell in row:
                html_table += f"{cell}"
            html_table += "\n"
        
        # Create formatted HTML text with style
        formatted_html = f"{html_table}"
        self.conn.send_result(formatted_html,style=style)


    def list_args(self, spec :BuiltinFunctionDefinition):
        table_data, style = self.list_args_table(spec)

        html_table = ""
        for row in table_data:
            html_table += "\t"
            for cell in row:
                html_table += f"{cell}"
            html_table += "\n"
        
        self.conn.send_result(html_table, style=style)
    

    def run(self, check_until:Callable[[angr.SimulationManager],StopReason] = lambda _:StopReason.NONE, exclude:Callable[[angr.SimState],bool] = lambda _:False, single_step:bool = False):
        u = check_until
        exclusions = self.exclusions
        def check(simgr:angr.SimulationManager):
            r = u(simgr)
            if r != StopReason.NONE:
                return r
            state = simgr.one_active

            if self.breakpoints.filter(state):
                return StopReason.BREAKPOINT
            return StopReason.NONE

        def _exclude(state:angr.SimState)->bool:
            if exclude(state):
                return True
            return any([f.filter(state) for f in exclusions])
        
        self._run(self, check,_exclude, single=single_step)


import claripy
from dAngr.cli.grammar.execution_context import Variable
from dAngr.cli.grammar.expressions import ReferenceObject
from dAngr.cli.models import Register
from dAngr.utils import  SymBitVector, AngrType, Endness
from .base import BaseCommand


class MemoryCommands(BaseCommand):
    def __init__(self, debugger_core):
        super().__init__(debugger_core)


    def assign(self, target:ReferenceObject, value:int|bytes|str|SymBitVector|Variable):
        """
        Assign a value to some target symbol

        Args:
            target (ReferenceObject): Name of the symbol. Prefix with &reg for register, &mem for memory, &sym for symbol.
            value (int|bytes|str|SymBitVector|Variable): Value to set. Either a value or a prefixed register, memory or symbol.
        
        Short name: as
        
        """
        target.set_value(self.debugger.context,self.to_value(value))
        self.send_info(f"Value {value} assigned to {target}.")


    def add_to_stack(self, value:int|bytes|str|SymBitVector|Variable):
        """
        Add a value to the stack.

        Args:
            value (int|bytes|str|SymBitVector|Variable): Value to set at the stack.
        
        Short name: ast
        
        """
        self.debugger.add_to_stack(self.get_angr_value(value))
        self.send_info(f"Value {value} added to the stack.")

    def get_stack(self, length:int, offset:int=0):
        """
        Get the stack values.

        Args:
            length (int): Length of the stack.
            offset (int): Offset from the current stack pointer. Default is 0.

        Short name: gst
        
        """
        return self.debugger.get_stack(length, offset)

    def get_memory_string(self, address:int):
        """
        Get the memory value at a specific address.

        Args:
            address (int): Address in the memory
        
        Short name: mgs
        
        """
        return self.debugger.get_string_memory(address)
        
    def get_memory(self, address:int|SymBitVector, size:int|SymBitVector, endness:Endness=Endness.DEFAULT):
        """
        Get the memory value at a specific address.
        Supported Conversion Types: int, bytes, bool, double, hex, none.

        Args:
            address (int): Address in the memory
            size (int): Size of the memory
            endness (Endness): Endianness of the value.
        
        Short name: mg
        
        """
        return self.debugger.get_memory(address, size, endness=endness)
    
    def get_addr_for_name(self, name: str):
        """
        Get the address for a symbol name.
        Requires the options REVERSE_MEMORY_NAME_MAP and TRACK_ACTION_HISTORY in the entry state. 

        Args:
            name (str): Symbol name
        
        Short name: gan
        
        """
        addresses = []
        for addr in self.debugger.get_addr_for_symbol(name):
            addresses.append(addr)
        return addresses

    def get_stdin_variables(self):
        """
        Get the stdin variable.

        Short name: gsv
        
        """
        stdin_vars = []
        for _, symbol in self.debugger.get_stdin_variables():
            stdin_vars.append(symbol)
        return stdin_vars
    
    def set_memory(self, address:int|SymBitVector, value:AngrType, size:int|None=None, endness:Endness=Endness.DEFAULT):
        """
        Set a memory value at a specific address.
        Supported Conversion Types: int, bytes, bool, double, hex, none.

        Args:
            address (int|SymBitVector): Address in the memory
            value (AngrType): Value to set at the address.
            size (int|None): Size of the memory, default is None.
            endness (Endness): Endianness of the value.

        Short name: ms
        """
        value = self.get_angr_value(value)
        self.debugger.set_memory(address, value, size=size, endness=endness)        
        if isinstance(address, SymBitVector):
            a = str(address)
        else:
            a = hex(address)
        self.send_info(f"Memory at {a}: {value}.")
    
    def add_static_pointer(self, name:str, value:int, size_bytes:int = 4):
        """
        Create a static pointer variable.

        Args:
            name (str): Name of the pointer
            value (int): pointer value.
            size_bytes (int): size of the variable in bytes, default 4.
        
        Short name: asp
        
        """
        val = claripy.BVV(value, size_bytes*8)
        self.debugger.context.add_variable(name, val)
        self.send_info(f"Pointer {name}: {value} added to variables.")
    
    def set_register(self, name:str, value:int|SymBitVector|Variable):
        """
        Set a register value. Same as $reg.{name} = {value}.

        Args:
            name (str): Register name
            value (int|SymBitVector|Variable): Value to set at the register.
        
        Short name: rs
        
        """
        self.debugger.set_register(name, self.to_value(value)) # type: ignore
        self.send_info(f"Register {name}: {hex(value) if isinstance(value, int) else value}.")
    
    # def set_register_to_symbol(self, name:str, value:Variable):
    #     """
    #     Set a register value to a symbol.

    #     Args:
    #         name (str): Register name
    #         value (Variable): Symbol name to set at the register.
        
    #     Short name: rss
        
    #     """
        
    #     v = self.debugger.get_new_symbol_object(value)
    #     if isinstance(v, claripy.ast.FP) or isinstance(v, claripy.ast.String):
    #         raise ValueError("Symbol cannot be a floating point or string value.")
    #     self.debugger.set_register(name, v)
    #     self.send_info(f"Register {name}: {value}.")

    def get_register(self, name:str):
        """
        Get a register value, same as $reg.{name}.

        Args:
            name (str): Register name
        
        Short name: rg
        
        """
        return self.debugger.get_register(name)

    def list_registers(self):
        """
        List all the registers.

        Short name: rl
        
        """
        regs=[]
        registers = self.debugger.list_registers()
        for reg, (offset, size) in registers.items():
            # Reading register value; size is in bits, need to convert to bytes
            value = self.debugger.get_register_value(reg)
            regs.append(Register(reg,size,offset,value))
        
        return regs

    def unconstrained_fill(self, symbolic:bool=False):
        """
        Fill the memory and registers with symbolic values.

        Args:
            symbolic (bool): fill with symbolic values, else with zeros. Default is zeros.
        
        Short name: uf
        
        """
        self.debugger.unconstrained_fill(symbolic)
        self.send_info( f"Fill {'with symbols' if symbolic else 'with zeros'}.")



from dAngr.angr_ext.debugger import Debugger
from dAngr.cli.debugger_commands import BaseCommand
from dAngr.cli.debugger_commands.filters import FilterCommands
from dAngr.cli.filters import AddressFilter, AndFilterList, Filter, FilterFunction, FilterList, FunctionFilter, OrFilterList, SourceFilter, StdStreamFilter
from dAngr.exceptions import DebuggerCommandError, ExecutionError
from dAngr.utils.utils import StreamType

class ExclusionCommands(BaseCommand):

    def __init__(self, debugger:Debugger):
        super().__init__(debugger)

    def add_exclusion(self, address:int):
        """
        Add an exclusion filter for a given address.

        Args:
            address (int): Address to exclude.
        
        Short name: exa
        """
        FilterCommands(self.debugger).filter(True, FilterCommands(self.debugger).by_address(address))

    def exclude(self, *filters:Filter):
        """
        Add a filter to the list of exclusions.

        Args:
            filters (tuple): The filters to add. Can be an address, source file and line number, function name or stream text.
        
        Short name: exaf
        """
        FilterCommands(self.debugger).filter(True, *filters)

    
    def remove_exclusion_filter(self, index:int):
        """
        Remove a filter from the list of exclusions.

        Args:
            index (int): Index of the filter found using list_filters.
        
        Short name: exr
        """
        FilterCommands(self.debugger).remove_filter(index, True)
    

    
    def enable_exclusion(self, index:int=0, enable:bool=True):
        """
        Enable filter at given index.

        Args:
            index (int): Index of the filter found using list_exclusions.
            enable (bool): True to enable, False to disable.
        
        Short name: exe
        """
        list = self.debugger.exclusions
        if index >= len(list):
            raise DebuggerCommandError(f"Index {index} out of range.")
        list[index].enabled = enable
        self.send_info(f"Exclusion filter {'enabled' if enable else 'disabled'}.")

    
    def list_exclusions(self):
        """
        List all exclusions.
        
        Short name: exl
        """
        list = self.debugger.exclusions
        if len(list) == 0:
            self.send_info(f'No exclusions found.')
            return []
        return f'Exclusion(s): {"\n\t".join([f"[{i}] {b}" for i,b in enumerate(list)])}'
    
    
    def clear_exclusions(self):
        """
        Clear all exclusions.

        Short name: fc
        """
        self.debugger.exclusions.clear()
        self.send_info( "All exclusions cleared.")




from typing import Any, Dict
import angr
from dAngr.utils import Variable
from dAngr.exceptions import KeyError


class ExecutionContext:
    def __init__(self, parent=None):
        from .definitions import Definition
        self._variables: Dict[str,Variable] = {}
        self._definitions:Dict[str,Definition] = {}
        self._parent:ExecutionContext|None= parent # type: ignore
        self.enums = {"options":angr.options}
        self.python_context = {}
        self.return_value = None
        
    @property
    def root(self):
        return self._parent.root if self._parent else self
    
    def clone(self):
        e = ExecutionContext() # type: ignore
        e._variables = self._variables.copy()
        e._definitions = self._definitions.copy()
        e._parent = self
        return e
    
    def __getitem__(self, name) -> Variable:
        if name in self._variables:
            return self._variables[name]
        if self._parent and name in self._parent.variables:
            return self._parent[name]
        raise KeyError(f"Unknown variable: {name}")
    
    def find_variable(self, name:str):
        if name in self._variables:
            return self._variables[name]
        if self._parent and name in self._parent.variables:
            return self._parent[name]
        if "@"+name in self.root.variables:
            return self.root.variables["@"+name]
        return None
    
    def find_enum(self, name:str):
        if name in self.enums:
            return self.enums[name]
        if self._parent:
            return self._parent.find_enum(name)
        return None
    
    
    def __setitem__(self, name:str, value):
        assert not isinstance(value, Variable)
        if name in self._variables:
            self._variables[name]._value = value
        # Check if itself function context is
        elif self._parent and name in self._parent.variables:
            self._parent[name].value = value
        else:
            self.add_variable(name, value)

    def add_variable(self, name:str, value:Any):
        assert not isinstance(value, Variable)
        self._variables[name] = Variable(name, value)

    def remove_variable(self, name:str):
        if name in self._variables:
            del self._variables[name]
    def add_definition(self, name:str, value):
        self._definitions[name] = value # type: ignore
    def remove_definition(self, name:str):
        if name in self._definitions:
            del self._definitions[name]

    def get_definition(self, name:str):
        if definition := self.find_definition(name):
            return definition
        raise KeyError(f"Unknown definition: {name}")
    def find_definition(self, name:str):
        if name in self._definitions:
            return self._definitions[name]
        if self._parent and name in self._parent.definitions:
            return self._parent.get_definition(name)
        return None
    @property
    def definitions(self):
        # merge dict from the parent with precedence to the current context
        return {**self._parent.definitions, **self._definitions} if self._parent else self._definitions
    @property
    def functions(self):
        from .definitions import FunctionDefinition
        return {k:v for k,v in self.definitions.items() if isinstance(v, FunctionDefinition)}
    
    def find_function(self, package:str|None, name:str):
        from dAngr.cli.command_line_debugger import BuiltinFunctionDefinition
        if f:= next((f for f in self.functions.values() if (f.name == name ) and (f.package == package if package else True)), None):
            return f
        else:
            return next((f for f in self.functions.values() if isinstance(f, BuiltinFunctionDefinition) and (f.short_name == name ) and not package), None)
    
    @property
    def variables(self)->Dict[str,Variable]:
        # merge dict from the parent with precedence to the current context
        return {**self._parent.variables, **self._variables} if self._parent else self._variables
